library deeper;

import 'dart:convert';

dynamic maps(
  dynamic src,
  String key,
  dynamic Function(dynamic value, List<dynamic> keyToValue) mapper, {
  List<dynamic> keyParents = const [],
  bool onKeyNotExistIgnore = true,
}) {
  dynamic mapsRecursive(
    dynamic src,
    String key,
    dynamic Function(dynamic value, List<dynamic> keyToValue) mapper, {
    List<dynamic> keyParents = const [],
    bool onKeyNotExistIgnore = true,
  }) {
    var keys = key.split('.');
    var first = keys.first;
    var tail = keys.sublist(1);
    try {
      dynamic mapSingleField(dynamic field, k) => mapper(field, keyParents + [k]);
      dynamic mapMultipleField(dynamic fields) => fields.asMap().entries.map((entry) {
            var idx = entry.key;
            return mapSingleField(entry.value, keyParents + [idx]);
          });

      if (first != '[]' && !src.containsKey(first) || first == '[]' && !(src is List)) {
        if (onKeyNotExistIgnore) {
          return src;
        } else {
          throw Exception('onKeyNotExistThrowFormatException');
        }
      }

      if (tail.isEmpty) {
        if (first == '[]') {
          src = mapMultipleField(src);
        } else {
          src[first] = mapSingleField(src[first], first);
        }
      } else {
        if (first == '[]') {
          src = src.asMap().entries.map((entry) {
            var idx = entry.key;
            var item = entry.value;
            return mapsRecursive(
              item,
              tail.join('.'),
              mapper,
              keyParents: keyParents + [idx as int],
              onKeyNotExistIgnore: onKeyNotExistIgnore,
            );
          }).toList();
        } else {
          src[first] = mapsRecursive(
            src[first],
            tail.join('.'),
            mapper,
            keyParents: keyParents + [first],
            onKeyNotExistIgnore: onKeyNotExistIgnore,
          );
        }
      }
    } catch (e) {
      if (e is FormatException) rethrow;
      throw FormatException('Key: ${(keyParents + [first]).join('.')} not exist!');
    }
    return src;
  }

  return mapsRecursive(
    json.decode(json.encode(src)),
    key,
    mapper,
    keyParents: keyParents,
    onKeyNotExistIgnore: onKeyNotExistIgnore,
  );
}
